/*-
 * Copyright 2017 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"

#if !HAVE_PCRE2_8 && !HAVE_PCRE2_16 && !HAVE_PCRE2_32
#error "none of the 8-, 16- or 32-bit pcre2 libs specified in config.h"
#endif

#define PCRE2_CODE_UNIT_WIDTH 0
#include "pcre2.h"

#if HAVE_PCRE2_8
#define LIB_SMALLEST (LIB8)
#elif HAVE_PCRE2_16
#define LIB_SMALLEST (LIB16)
#elif HAVE_PCRE2_32
#define LIB_SMALLEST (LIB32)
#endif

typedef union {
#if HAVE_PCRE2_8
	pcre2_code_8 *code8;
#endif
#if HAVE_PCRE2_16
	pcre2_code_16 *code16;
#endif
#if HAVE_PCRE2_32
	pcre2_code_32 *code32;
#endif
} pcre2_code_t;

typedef union {
#if HAVE_PCRE2_8
	pcre2_compile_context_8 *ccontext8;
#endif
#if HAVE_PCRE2_16
	pcre2_compile_context_16 *ccontext16;
#endif
#if HAVE_PCRE2_32
	pcre2_compile_context_32 *ccontext32;
#endif
} pcre2_compile_context_t;

typedef struct {
#if HAVE_PCRE2_8
	pcre2_general_context_8 *gcontext8;
#endif
#if HAVE_PCRE2_16
	pcre2_general_context_16 *gcontext16;
#endif
#if HAVE_PCRE2_32
	pcre2_general_context_32 *gcontext32;
#endif
} pcre2_general_context_t;

typedef union {
#if HAVE_PCRE2_8
	pcre2_match_context_8 *mcontext8;
#endif
#if HAVE_PCRE2_16
	pcre2_match_context_16 *mcontext16;
#endif
#if HAVE_PCRE2_32
	pcre2_match_context_32 *mcontext32;
#endif
} pcre2_match_context_t;

typedef union {
#if HAVE_PCRE2_8
	pcre2_match_data_8 *mdata8;
#endif
#if HAVE_PCRE2_16
	pcre2_match_data_16 *mdata16;
#endif
#if HAVE_PCRE2_32
	pcre2_match_data_32 *mdata32;
#endif
} pcre2_match_data_t;

static const int have_lib[_LIB_MAX] = {
	[LIB8] = HAVE_PCRE2_8, [LIB16] = HAVE_PCRE2_16, [LIB32] = HAVE_PCRE2_32
};

static inline int
PCRE2_COMPILE_CONTEXT_CREATE(enum code_size codesz,
			     pcre2_compile_context_t *cctx)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		cctx->ccontext8 = pcre2_compile_context_create_8(NULL);
		return cctx->ccontext8 == NULL;
		break;
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		cctx->ccontext16 = pcre2_compile_context_create_16(NULL);
		return cctx->ccontext16 == NULL;
		break;
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		cctx->ccontext32 = pcre2_compile_context_create_32(NULL);
		return cctx->ccontext32 == NULL;
		break;
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline void
PCRE2_COMPILE_CONTEXT_FREE(enum code_size codesz, pcre2_compile_context_t *cctx)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		pcre2_compile_context_free_8(cctx->ccontext8);
		break;
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		pcre2_compile_context_free_16(cctx->ccontext16);
		break;
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		pcre2_compile_context_free_32(cctx->ccontext32);
		break;
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_COMPILE(enum code_size codesz, pcre2_code_t *code, VCL_STRING pattern,
              PCRE2_SIZE length, uint32_t options, int *err_code,
              PCRE2_SIZE *err_offset, pcre2_compile_context_t *cctx)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		code->code8 = pcre2_compile_8((PCRE2_SPTR8)pattern, length,
			                      options, err_code, err_offset,
			                      cctx->ccontext8);
		return code->code8 == NULL;
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		code->code16 = pcre2_compile_16((PCRE2_SPTR16)pattern, length,
			                        options, err_code, err_offset,
			                        cctx->ccontext16);
		return code->code16 == NULL;
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		code->code32 = pcre2_compile_32((PCRE2_SPTR32)pattern, length,
			                        options, err_code, err_offset,
			                        cctx->ccontext32);
		return code->code32 == NULL;
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_JIT_COMPILE(enum code_size codesz, pcre2_code_t *code, uint32_t options)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_jit_compile_8(code->code8, options);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_jit_compile_16(code->code16, options);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_jit_compile_32(code->code32, options);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline void
PCRE2_CODE_FREE(enum code_size codesz, pcre2_code_t *code)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		pcre2_code_free_8(code->code8);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		pcre2_code_free_16(code->code16);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		pcre2_code_free_32(code->code32);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_MATCH(enum code_size codesz, pcre2_code_t *code, VCL_STRING subject,
	    PCRE2_SIZE length, PCRE2_SIZE start_offset, uint32_t options,
	    pcre2_match_data_t *mdata, pcre2_match_context_t *mctx)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_match_8(code->code8, (PCRE2_SPTR8)subject,
			             length, start_offset, options,
				     mdata->mdata8, mctx->mcontext8);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_match_16(code->code16, (PCRE2_SPTR16)subject,
				      length, start_offset, options,
				      mdata->mdata16, mctx->mcontext16);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_match_32(code->code32, (PCRE2_SPTR32)subject,
				      length, start_offset, options,
				      mdata->mdata32, mctx->mcontext32);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_GENERAL_CONTEXT_CREATE(pcre2_general_context_t *gctx,
			     void *(*private_malloc)(PCRE2_SIZE, void *),
			     void (*private_free)(void *, void *), void *ws)
{
#if HAVE_PCRE2_8
	VSL(SLT_Debug, 0, "gctx_create8 gctx = %p, malloc = %p, free = %p",
	    gctx, private_malloc, private_free);
	gctx->gcontext8 = pcre2_general_context_create_8(private_malloc,
							 private_free, ws);
	VSL(SLT_Debug, 0, "gctx_create8 returned gcontext8 = %p",
	    gctx->gcontext8);
	if (gctx->gcontext8 == NULL)
		return -1;
#endif
#if HAVE_PCRE2_16
	VSL(SLT_Debug, 0, "gctx_create16 gctx = %p, malloc = %p, free = %p",
	    gctx, private_malloc, private_free);
	gctx->gcontext16 = pcre2_general_context_create_16(private_malloc,
							   private_free, ws);
	VSL(SLT_Debug, 0, "gctx_create16 returned gcontext16 = %p",
	    gctx->gcontext16);
	if (gctx->gcontext16 == NULL)
		return -1;
#endif
#if HAVE_PCRE2_32
	VSL(SLT_Debug, 0, "gctx_create32 gctx = %p, malloc = %p, free = %p",
	    gctx, private_malloc, private_free);
	gctx->gcontext32 = pcre2_general_context_create_32(private_malloc,
							   private_free, ws);
	VSL(SLT_Debug, 0, "gctx_create32 returned gcontext32 = %p",
	    gctx->gcontext32);
	if (gctx->gcontext32 == NULL)
		return -1;
#endif
	return 0;
}

static inline int
PCRE2_MATCH_CONTEXT_CREATE(enum code_size codesz, pcre2_match_context_t *mctx,
			   const pcre2_general_context_t *gctx)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		mctx->mcontext8 = pcre2_match_context_create_8(gctx->gcontext8);
		VSL(SLT_Debug, 0, "mctx_create returned mcontext8 = %p",
		    mctx->mcontext8);
		return mctx->mcontext8 == NULL;
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		mctx->mcontext16
			= pcre2_match_context_create_16(gctx->gcontext16);
		return mctx->mcontext16 == NULL;
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		mctx->mcontext32
			= pcre2_match_context_create_32(gctx->gcontext32);
		return mctx->mcontext32 == NULL;
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline void
PCRE2_MATCH_CONTEXT_FREE(enum code_size codesz, pcre2_match_context_t *mctx)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		pcre2_match_context_free_8(mctx->mcontext8);
		break;
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		pcre2_match_context_free_16(mctx->mcontext16);
		break;
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		pcre2_match_context_free_32(mctx->mcontext32);
		break;
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_MATCH_DATA_CREATE_FROM_PATTERN(enum code_size codesz,
				     pcre2_match_data_t *mdata,
				     pcre2_code_t *code,
				     pcre2_general_context_t *gctx)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		mdata->mdata8
			= pcre2_match_data_create_from_pattern_8(code->code8,
							       gctx->gcontext8);
		return mdata->mdata8 == NULL;
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		mdata->mdata16
			= pcre2_match_data_create_from_pattern_16(code->code16,
							      gctx->gcontext16);
		return mdata->mdata16 == NULL;
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		mdata->mdata32
			= pcre2_match_data_create_from_pattern_32(code->code32,
							      gctx->gcontext32);
		return mdata->mdata8 == NULL;
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_GET_ERROR_MESSAGE(enum code_size codesz, int errorcode, char *buf,
			unsigned bytes)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_get_error_message_8(errorcode,
						 (PCRE2_UCHAR8 *)buf, bytes)
			+ 1;
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_get_error_message_16(errorcode,
						  (PCRE2_UCHAR16 *)buf,
						  bytes / 2) * 2 + 1;
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_get_error_message_32(errorcode,
						  (PCRE2_UCHAR32 *)buf,
						  bytes / 4) * 4 + 1;
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_SET_BSR(enum code_size codesz, pcre2_compile_context_t *cctx,
	      uint32_t value)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_set_bsr_8(cctx->ccontext8, value);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_set_bsr_16(cctx->ccontext16, value);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_set_bsr_32(cctx->ccontext32, value);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_SET_MAX_PATTERN_LENGTH(enum code_size codesz,
			     pcre2_compile_context_t *cctx, uint32_t value)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_set_max_pattern_length_8(cctx->ccontext8, value);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_set_max_pattern_length_16(cctx->ccontext16, value);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_set_max_pattern_length_32(cctx->ccontext32, value);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_SET_NEWLINE(enum code_size codesz, pcre2_compile_context_t *cctx,
		  uint32_t value)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_set_newline_8(cctx->ccontext8, value);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_set_newline_16(cctx->ccontext16, value);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_set_newline_32(cctx->ccontext32, value);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_SET_PARENS_NEST_LIMIT(enum code_size codesz,
			    pcre2_compile_context_t *cctx, uint32_t value)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_set_parens_nest_limit_8(cctx->ccontext8, value);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_set_parens_nest_limit_16(cctx->ccontext16, value);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_set_parens_nest_limit_32(cctx->ccontext32, value);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_SET_MATCH_LIMIT(enum code_size codesz, pcre2_match_context_t *mctx,
		      uint32_t value)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_set_match_limit_8(mctx->mcontext8, value);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_set_match_limit_16(mctx->mcontext16, value);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_set_match_limit_32(mctx->mcontext32, value);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_SET_OFFSET_LIMIT(enum code_size codesz, pcre2_match_context_t *mctx,
		       uint32_t value)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_set_offset_limit_8(mctx->mcontext8, value);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_set_offset_limit_16(mctx->mcontext16, value);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_set_offset_limit_32(mctx->mcontext32, value);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_SET_RECURSION_LIMIT(enum code_size codesz, pcre2_match_context_t *mctx,
			  uint32_t value)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_set_recursion_limit_8(mctx->mcontext8, value);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_set_recursion_limit_16(mctx->mcontext16, value);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_set_recursion_limit_32(mctx->mcontext32, value);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}

static inline int
PCRE2_CONFIG(enum code_size codesz, uint32_t what, void *where)
{
	switch(codesz) {
	case LIB8:
#if HAVE_PCRE2_8
		return pcre2_config_8(what, where);
#else
		WRONG("8-bit lib not supported");
#endif
	case LIB16:
#if HAVE_PCRE2_16
		return pcre2_config_16(what, where);
#else
		WRONG("16-bit lib not supported");
#endif
	case LIB32:
#if HAVE_PCRE2_32
		return pcre2_config_32(what, where);
#else
		WRONG("32-bit lib not supported");
#endif
	default:
		WRONG("Illegal code size");
	}
}
