/*-
 * Copyright 2017 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "config.h"

#include <stdint.h>
#include <string.h>

#include "vcl.h"
#include "vrt.h"
#include "vas.h"
#include "vdef.h"
#include "cache/cache.h"

#include "vcc_if.h"

/* XXX enum in a separate .h? */
enum code_size {
	LIB8 = 0,
	LIB16,
	LIB32,
	_LIB_MAX
};

#include "pcre2_compat.h"

#define ERR(ctx, msg) \
        errmsg((ctx), "vmod pcre2 error: " msg)

#define VERR(ctx, fmt, ...) \
        errmsg((ctx), "vmod pcre2 error: " fmt, __VA_ARGS__)

#define VERRNOMEM(ctx, fmt, ...) \
        VERR((ctx), fmt ", out of space", __VA_ARGS__)

#define ERRNOMEM(ctx, msg) \
        ERR((ctx), msg ", out of space")

#define INIT(ctx) (((ctx)->method & VCL_MET_INIT) != 0)

struct vmod_pcre2_regex {
	unsigned	magic;
#define VMOD_PCRE2_REGEX_MAGIC 0x3adb2a78
	pcre2_code_t	code;
	char		*vcl_name;
	enum code_size	codesz;
};

/* PRIV_CALL scope for match */
struct match_call {
	unsigned		magic;
#define VMOD_PCRE2_MATCH_CALL_MAGIC 0x60e5bd33
	pcre2_match_context_t	mctx;
	uint32_t		options;
	enum code_size		codesz;
};

static const pcre2_general_context_t gctx_malloc = { NULL };
static const int code_size[_LIB_MAX] = {
	[LIB8] = 8, [LIB16] = 16, [LIB32] = 32
};
static int have_jit[_LIB_MAX] = { 0 };

void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (ctx->method == VCL_MET_INIT) {
		AN(ctx->msg);
		VSB_vprintf(ctx->msg, fmt, args);
		VRT_handling(ctx, VCL_RET_FAIL);
	}
	else if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		/* Should this ever happen in vcl_fini() ... */
		VSL(SLT_VCL_Error, 0, fmt, args);
	va_end(args);
}

static inline void
WS_Allocated(struct ws * const restrict ws, const void * const restrict ptr,
	     const size_t len)
{
        assert((char *)ptr >= ws->s && ((char *)ptr + len) <= ws->f);
}

static inline enum code_size
decode_lib_enum(VCL_ENUM libs)
{
	if (libs == NULL)
		return LIB_SMALLEST;
	else if (strcmp(libs, "LIB8") == 0)
		return LIB8;
	else if (strcmp(libs, "LIB16") == 0)
		return LIB16;
	else if (strcmp(libs, "LIB32") == 0)
		return LIB32;
	else
		WRONG("Illegal lib enum");
}

/* Event function */

int __match_proto__(vmod_event_f)
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
        (void) ctx;
        (void) priv;

        if (e == VCL_EVENT_LOAD) {
		int ret;
		uint32_t where;

		if (have_lib[LIB8]) {
			ret = PCRE2_CONFIG(LIB8, PCRE2_CONFIG_JIT, &where);
			assert(ret >= 0);
			have_jit[LIB8] = where;
		}
		if (have_lib[LIB16]) {
			ret = PCRE2_CONFIG(LIB16, PCRE2_CONFIG_JIT, &where);
			assert(ret >= 0);
			have_jit[LIB16] = where;
		}
		if (have_lib[LIB32]) {
			ret = PCRE2_CONFIG(LIB32, PCRE2_CONFIG_JIT, &where);
			assert(ret >= 0);
			have_jit[LIB32] = where;
		}
	}
        return 0;
}

/* Object regex */

static void *
ws_malloc(PCRE2_SIZE sz, void *ws)
{
	void *p;

	VSL(SLT_Debug, 0, "sz = %zu, ws = %p", sz, ws);
	WS_Assert(ws);
	VSL(SLT_Debug, 0, "WS_Asserted");
	//return WS_Alloc(ws, sz);
	p = WS_Alloc(ws, sz);
	VSL(SLT_Debug, 0, "WS_Alloced %p", p);
	return p;
}

static void
ws_free(void *ptr, void *ws)
{
	VSL(SLT_Debug, 0, "free ptr = %p, ws = %p", ptr, ws);
	WS_Assert(ws);
	WS_Allocated(ws, ptr, 0);
}

static void
match_call_free(void *priv)
{
	struct match_call *mctx_call;

	if (priv == NULL)
		return;
	CAST_OBJ(mctx_call, priv, VMOD_PCRE2_MATCH_CALL_MAGIC);
	PCRE2_MATCH_CONTEXT_FREE(mctx_call->codesz, &mctx_call->mctx);
	FREE_OBJ(mctx_call);
}

VCL_VOID
vmod_regex__init(VRT_CTX, struct vmod_pcre2_regex **regexp,
		 const char *vcl_name, VCL_STRING pattern, VCL_ENUM libs,
		 VCL_BOOL allow_empty_class, VCL_BOOL anchored, VCL_ENUM bsrs,
		 VCL_BOOL alt_bsux, VCL_BOOL alt_circumflex,
		 VCL_BOOL alt_verbnames, VCL_BOOL caseless,
		 VCL_BOOL dollar_endonly, VCL_BOOL dotall, VCL_BOOL dupnames,
		 VCL_BOOL extended, VCL_BOOL firstline, VCL_STRING locale,
		 VCL_BOOL match_unset_backref, VCL_INT max_pattern_len,
		 VCL_BOOL multiline, VCL_BOOL never_backslash_c,
		 VCL_BOOL never_ucp, VCL_BOOL never_utf, VCL_ENUM newlines,
		 VCL_BOOL no_auto_capture, VCL_BOOL no_auto_possess,
		 VCL_BOOL no_dotstar_anchor, VCL_BOOL no_start_optimize,
		 VCL_BOOL no_utf_check, VCL_INT parens_nest_limit,
		 VCL_BOOL ucp, VCL_BOOL ungreedy, VCL_BOOL use_offset_limit,
		 VCL_BOOL utf)
{
	struct vmod_pcre2_regex *regex;
	enum code_size codesz;
	pcre2_code_t code;
	pcre2_compile_context_t ccontext;
	uint32_t options = 0, val;
	int ret, err_code = 0;
	PCRE2_SIZE err_offset;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(regexp);
	AZ(*regexp);
	AN(vcl_name);
	if (pattern == NULL) {
		VERR(ctx, "pattern is NULL in %s constructor", vcl_name);
		return;
	}
	codesz = decode_lib_enum(libs);
	if (!have_lib[codesz]) {
		VERR(ctx, "%d-bit library not supported in %s constructor",
		     code_size[codesz], vcl_name);
		return;
	}
	ret = PCRE2_COMPILE_CONTEXT_CREATE(codesz, &ccontext);
	if (ret != 0) {
		VERR(ctx, "failed to create compile context in %s constructor",
		     vcl_name);
		return;
	}

	if (bsrs != NULL) {
		if (strcmp("ANYCRLF", bsrs) == 0)
			val = PCRE2_BSR_ANYCRLF;
		else if (strcmp("UNICODE", bsrs) == 0)
			val = PCRE2_BSR_UNICODE;
		else
			WRONG("Illegal bsr enum value");
		ret = PCRE2_SET_BSR(codesz, &ccontext, val);
		AZ(ret);
	}
	if (locale != NULL) {
		VERR(ctx, "locales not implemented yet in %s constructor",
		     vcl_name);
		return;
	}
	if (max_pattern_len != 0) {
		ret = PCRE2_SET_MAX_PATTERN_LENGTH(codesz, &ccontext,
						   (PCRE2_SIZE)max_pattern_len);
		AZ(ret);
	}
	if (newlines != NULL) {
		if (strcmp("CR", newlines) == 0)
			val = PCRE2_NEWLINE_CR;
		else if (strcmp("LF", newlines) == 0)
			val = PCRE2_NEWLINE_LF;
		else if (strcmp("CRLF", newlines) == 0)
			val = PCRE2_NEWLINE_CRLF;
		else if (strcmp("ANYCRLF", newlines) == 0)
			val = PCRE2_NEWLINE_ANYCRLF;
		else if (strcmp("ANY", newlines) == 0)
			val = PCRE2_NEWLINE_ANY;
		else
			WRONG("Illegal newline enum value");
		ret = PCRE2_SET_NEWLINE(codesz, &ccontext, val);
		AZ(ret);
	}
	if (parens_nest_limit != 0) {
		ret = PCRE2_SET_PARENS_NEST_LIMIT(codesz, &ccontext,
						  (uint32_t)parens_nest_limit);
		AZ(ret);
	}

	/* XXX check illegal combinations such as never_ucp && ucp ... ? */
	if (anchored)
		options |= PCRE2_ANCHORED;
	if (allow_empty_class)
		options |= PCRE2_ALLOW_EMPTY_CLASS;
	if (alt_bsux)
		options |= PCRE2_ALT_BSUX;
	if (alt_circumflex)
		options |= PCRE2_ALT_CIRCUMFLEX;
	if (alt_verbnames)
		options |= PCRE2_ALT_VERBNAMES;
	if (caseless)
		options |= PCRE2_CASELESS;
	if (dollar_endonly)
		options |= PCRE2_DOLLAR_ENDONLY;
	if (dotall)
		options |= PCRE2_DOTALL;
	if (dupnames)
		options |= PCRE2_DUPNAMES;
	if (extended)
		options |= PCRE2_EXTENDED;
	if (firstline)
		options |= PCRE2_FIRSTLINE;
	if (match_unset_backref)
		options |= PCRE2_MATCH_UNSET_BACKREF;
	if (multiline)
		options |= PCRE2_MULTILINE;
	if (never_backslash_c)
		options |= PCRE2_NEVER_BACKSLASH_C;
	if (never_ucp)
		options |= PCRE2_NEVER_UCP;
	if (never_utf)
		options |= PCRE2_NEVER_UTF;
	if (no_auto_capture)
		options |= PCRE2_NO_AUTO_CAPTURE;
	if (no_auto_possess)
		options |= PCRE2_NO_AUTO_POSSESS;
	if (no_dotstar_anchor)
		options |= PCRE2_NO_DOTSTAR_ANCHOR;
	if (no_start_optimize)
		options |= PCRE2_NO_START_OPTIMIZE;
	if (no_utf_check)
		options |= PCRE2_NO_UTF_CHECK;
	if (ucp)
		options |= PCRE2_UCP;
	if (ungreedy)
		options |= PCRE2_UNGREEDY;
	if (use_offset_limit)
		options |= PCRE2_USE_OFFSET_LIMIT;
	if (utf)
		options |= PCRE2_UTF;

	/* XXX set the length via parameter */
	ret = PCRE2_COMPILE(codesz, &code, pattern, PCRE2_ZERO_TERMINATED,
			    options, &err_code, &err_offset, &ccontext);
	if (ret != 0) {
		char *buf = WS_Snapshot(ctx->ws);
		unsigned bytes = WS_Reserve(ctx->ws, 0);
		ret = PCRE2_GET_ERROR_MESSAGE(codesz, err_code, buf, bytes);
		if (ret == PCRE2_ERROR_BADDATA) {
			WS_ReleaseP(ctx->ws, buf);
			VERR(ctx, "Cannot compile '%s' in %s constructor "
			     "(unknown error code)", pattern, vcl_name);
		}
		else if (ret == PCRE2_ERROR_NOMEMORY) {
			WS_Release(ctx->ws, strlen(buf) + 1);
			VERR(ctx, "Cannot compile '%s' in %s constructor: %s"
			     "(truncated error message) at offset %zu", pattern,
			     vcl_name, err_offset);
		}
		else {
			WS_Release(ctx->ws, ret);
			VERR(ctx, "Cannot compile '%s' in %s constructor: %s "
			     "at offset %zu", pattern, vcl_name, buf,
			     err_offset);
		}
		WS_Reset(ctx->ws, buf);
		PCRE2_COMPILE_CONTEXT_FREE(codesz, &ccontext);
		return;
	}
	if (have_jit[codesz]) {
		/* XXX check option compatibility; disable via param */
		/* XXX set complete or soft/hard partial via param */
		options |= PCRE2_JIT_COMPLETE;
		ret = PCRE2_JIT_COMPILE(codesz, &code, options);
		if (ret != 0) {
			/* XXX DRY */
			char *buf = WS_Snapshot(ctx->ws);
			unsigned bytes = WS_Reserve(ctx->ws, 0);
			ret = PCRE2_GET_ERROR_MESSAGE(codesz, ret, buf, bytes);
			if (ret == PCRE2_ERROR_BADDATA) {
				WS_ReleaseP(ctx->ws, buf);
				VERR(ctx, "Cannot jit-compile '%s' in %s "
				     "constructor (unknown error code)",
				     pattern, vcl_name);
			}
			else if (ret == PCRE2_ERROR_NOMEMORY) {
				WS_Release(ctx->ws, strlen(buf) + 1);
				VERR(ctx, "Cannot jit-compile '%s' in %s "
				     "constructor: %s (truncated error message)",
				     pattern, vcl_name);
			}
			else {
				WS_Release(ctx->ws, ret);
				VERR(ctx, "Cannot jit-compile '%s' in %s "
				     "constructor: %s ", pattern, vcl_name,
				     buf);
			}
			WS_Reset(ctx->ws, buf);
			PCRE2_COMPILE_CONTEXT_FREE(codesz, &ccontext);
			return;
		}
	}

	ALLOC_OBJ(regex, VMOD_PCRE2_REGEX_MAGIC);
	AN(regex);
	*regexp = regex;
	regex->code = code;
	regex->vcl_name = strdup(vcl_name);
	AN(regex->vcl_name);
	regex->codesz = codesz;
	PCRE2_COMPILE_CONTEXT_FREE(codesz, &ccontext);
}

VCL_VOID
vmod_regex__fini(struct vmod_pcre2_regex **regexp)
{
	struct vmod_pcre2_regex *regex;

	if (regexp == NULL)
		return;
	regex = *regexp;
	if (regex == NULL)
		return;
	CHECK_OBJ(regex, VMOD_PCRE2_REGEX_MAGIC);
	if (regex->vcl_name != NULL)
		free(regex->vcl_name);
	PCRE2_CODE_FREE(regex->codesz, &regex->code);
	FREE_OBJ(regex);
}

VCL_BOOL
vmod_regex_match(VRT_CTX, struct vmod_pcre2_regex *regex,
		 struct vmod_priv *mctx_call, struct vmod_priv *gctx_task,
		 VCL_STRING subject, VCL_INT len, VCL_BOOL anchored,
		 VCL_INT match_limit, VCL_INT offset_limit, VCL_BOOL notbol,
		 VCL_BOOL noteol, VCL_BOOL notempty, VCL_BOOL notempty_atstart,
		 VCL_BOOL no_jit, VCL_BOOL no_utf_check, VCL_ENUM partials,
		 VCL_INT recursion_limit)
{
	pcre2_general_context_t *gctx;
	pcre2_match_data_t *mdata;
	struct vmod_priv *match_task = NULL;
	struct match_call *match_opts;
	int ret;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(regex, VMOD_PCRE2_REGEX_MAGIC);
	assert(have_lib[regex->codesz]);
	AN(gctx_task);
	AN(mctx_call);
	if (subject == NULL) {
		VERR(ctx, "subject is NULL in %s.match()", regex->vcl_name);
		return 0;
	}

	/*
	 * The general context is task-scoped and shared by all objects
	 * from this VMOD in the task. It uses WS_Alloc as the custom
	 * allocator.
	 */
	if (gctx_task->priv == NULL) {
		if ((gctx = WS_Alloc(ctx->ws, sizeof(*gctx))) == NULL) {
			VERRNOMEM(ctx, "allocating general context in "
				  "%s.match()", regex->vcl_name);
			return 0;
		}
		VSLb(ctx->vsl, SLT_Debug, "ws_malloc = %p, ws_free = %p",
		     ws_malloc, ws_free);
		ret = PCRE2_GENERAL_CONTEXT_CREATE(gctx, ws_malloc, ws_free,
						   ctx->ws);
		if (ret != 0) {
			VERRNOMEM(ctx, "initializing general context in "
				  "%s.match()", regex->vcl_name);
			return 0;
		}
		gctx_task->priv = gctx;
		gctx_task->free = NULL;
	}
	else {
		WS_Allocated(ctx->ws, gctx_task->priv,
			     sizeof(pcre2_general_context_t));
		gctx = gctx_task->priv;
	}

	/* The match context and options are PRIV_CALL-scoped. */
	/* XXX param to decide whether these should be saved for the call */
	if (mctx_call->priv == NULL) {
		pcre2_match_context_t mctx;

		ret = PCRE2_MATCH_CONTEXT_CREATE(regex->codesz, &mctx, &gctx_malloc);
		if (ret != 0) {
			VERRNOMEM(ctx, "creating match context in %s.match()",
				  regex->vcl_name);
			return 0;
		}
		if (match_limit != 0) {
			ret = PCRE2_SET_MATCH_LIMIT(regex->codesz, &mctx,
						    (uint32_t)match_limit);
			AZ(ret);
		}
		if (offset_limit != 0) {
			ret = PCRE2_SET_OFFSET_LIMIT(regex->codesz, &mctx,
						     (uint32_t)offset_limit);
			AZ(ret);
		}
		if (recursion_limit != 0) {
			ret = PCRE2_SET_RECURSION_LIMIT(regex->codesz, &mctx,
						     (uint32_t)recursion_limit);
			AZ(ret);
		}
		ALLOC_OBJ(match_opts, VMOD_PCRE2_MATCH_CALL_MAGIC);
		AN(match_opts);
		match_opts->mctx = mctx;
		match_opts->codesz = regex->codesz;
		mctx_call->priv = match_opts;
		mctx_call->free = match_call_free;

		if (anchored != 0)
			match_opts->options |= PCRE2_ANCHORED;
		if (notbol != 0)
			match_opts->options |= PCRE2_NOTBOL;
		if (noteol != 0)
			match_opts->options |= PCRE2_NOTEOL;
		if (notempty != 0)
			match_opts->options |= PCRE2_NOTEMPTY;
		if (notempty_atstart != 0)
			match_opts->options |= PCRE2_NOTEMPTY_ATSTART;
		if (no_jit != 0)
			match_opts->options |= PCRE2_NO_JIT;
		if (no_utf_check != 0)
			match_opts->options |= PCRE2_NO_UTF_CHECK;

		if (partials != NULL) {
			VERR(ctx,
			   "Partial matching not implemented yet in %s.match()",
			     regex->vcl_name);
			return 0;
		}
	}
	else
		CAST_OBJ(match_opts, mctx_call->priv,
			 VMOD_PCRE2_MATCH_CALL_MAGIC);

	/* The match data block is task-scoped for this object only. */
	match_task = VRT_priv_task(ctx, regex);
	AN(match_task);
	if (match_task->priv == NULL) {
		if ((mdata = WS_Alloc(ctx->ws, sizeof(*mdata))) == NULL) {
			VERRNOMEM(ctx, "allocating match data in %s.match()",
				  regex->vcl_name);
			return 0;
		}
		ret = PCRE2_MATCH_DATA_CREATE_FROM_PATTERN(regex->codesz, mdata,
							   &regex->code, gctx);
		if (ret != 0) {
			VERRNOMEM(ctx, "initializing match data in "
				  "%s.match()", regex->vcl_name);
			return 0;
		}
		match_task->priv = mdata;
		match_task->free = NULL;
	}
	else {
		WS_Allocated(ctx->ws, match_task->priv,
			     sizeof(pcre2_match_context_t));
		mdata = match_task->priv;
	}

	/* XXX PRIV_CALL scope for options? */

	if (len == 0)
		len = PCRE2_ZERO_TERMINATED;
	else
		len *= code_size[regex->codesz] / 8;

	/* XXX param for start_offset */
	/* XXX free the match context? */
	ret = PCRE2_MATCH(regex->codesz, &regex->code, subject, (PCRE2_SIZE)len,
			  0, match_opts->options, mdata, &match_opts->mctx);
	if (ret == PCRE2_ERROR_NOMATCH)
		return 0;
	assert(ret != PCRE2_ERROR_PARTIAL); /* XXX */
	if (ret >= 0)
		return 1;

	/* XXX DRY */
	char *buf = WS_Snapshot(ctx->ws);
	unsigned bytes = WS_Reserve(ctx->ws, 0);
	ret = PCRE2_GET_ERROR_MESSAGE(regex->codesz, ret, buf, bytes);
	if (ret == PCRE2_ERROR_BADDATA) {
		WS_ReleaseP(ctx->ws, buf);
		VERR(ctx, "in %s.match() (unknown error code)",
		     regex->vcl_name);
	}
	else if (ret == PCRE2_ERROR_NOMEMORY) {
		WS_Release(ctx->ws, strlen(buf) + 1);
		VERR(ctx, "in %s.match(): %s (truncated error message)",
		     regex->vcl_name, buf);
	}
	else {
		WS_Release(ctx->ws, ret);
		VERR(ctx, "in %s.match(): %s ", regex->vcl_name, buf);
	}
	WS_Reset(ctx->ws, buf);
	return 0;
}

/* Functions */

VCL_BOOL
vmod_config_bool(VRT_CTX, VCL_ENUM configs, VCL_ENUM libs)
{
	uint32_t what, where;
	int ret;
	enum code_size lib;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	lib = decode_lib_enum(libs);
	if (!have_lib[lib]) {
		VERR(ctx, "%d-bit library not supported in pcre2.config_bool()",
		     code_size[lib]);
		return 0;
	}
	
	if (strcmp(configs, "JIT") == 0)
		what = PCRE2_CONFIG_JIT;
	else if (strcmp(configs, "STACKRECURSE") == 0)
		what = PCRE2_CONFIG_STACKRECURSE;
	else if (strcmp(configs, "UNICODE") == 0)
		what = PCRE2_CONFIG_UNICODE;
	else
		WRONG("Illegal config enum in config_bool");

	ret = PCRE2_CONFIG(lib, what, &where);
	assert(ret >= 0);
	return (VCL_BOOL) where;
}

VCL_STRING
vmod_config_str(VRT_CTX, VCL_ENUM configs, VCL_ENUM libs)
{
	uint32_t what;
	int ret;
	enum code_size lib;
	void *buf;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	lib = decode_lib_enum(libs);
	if (!have_lib[lib]) {
		VERR(ctx, "%d-bit library not supported in pcre2.config_str()",
		     code_size[lib]);
		return NULL;
	}

	if (strcmp(configs, "BSR") == 0)
		what = PCRE2_CONFIG_BSR;
	else if (strcmp(configs, "JITTARGET") == 0)
		what = PCRE2_CONFIG_JITTARGET;
	else if (strcmp(configs, "NEWLINE") == 0)
		what = PCRE2_CONFIG_NEWLINE;
	else if (strcmp(configs, "UNICODE_VERSION") == 0)
		what = PCRE2_CONFIG_UNICODE_VERSION;
	else if (strcmp(configs, "VERSION") == 0)
		what = PCRE2_CONFIG_VERSION;
	else
		WRONG("Illegal config enum in config_str");


	if (what == PCRE2_CONFIG_NEWLINE || what == PCRE2_CONFIG_BSR) {
		uint32_t where;

		ret = PCRE2_CONFIG(lib, what, &where);
		assert(ret >= 0);

		if (what == PCRE2_CONFIG_NEWLINE)
			switch(where) {
			case PCRE2_NEWLINE_CR:
				return "CR";
			case PCRE2_NEWLINE_LF:
				return "LF";
			case PCRE2_NEWLINE_CRLF:
				return "CRLF";
			case PCRE2_NEWLINE_ANY:
				return "ANY";
			case PCRE2_NEWLINE_ANYCRLF:
				return "ANYCRLF";
			default:
				WRONG("Illegal result for "
				      "PCRE2_CONFIG_NEWLINE");
			}

		assert (what == PCRE2_CONFIG_BSR);
		switch(where) {
		case PCRE2_BSR_UNICODE:
			return "UNICODE";
		case PCRE2_BSR_ANYCRLF:
			return "ANYCRLF";
		default:
			WRONG("Illegal result for PCRE2_CONFIG_BSR");
		}
	}

	ret = PCRE2_CONFIG(lib, what, NULL);
	if (what == PCRE2_CONFIG_JITTARGET && ret == PCRE2_ERROR_BADOPTION)
		return "JIT not supported";
	assert(ret > 0);
	ret *= (code_size[lib] / 8);
	if ((buf = WS_Alloc(ctx->ws, ret)) == NULL) {
		ERRNOMEM(ctx, "allocating result in pcre2.config_str()");
		return NULL;
	}
	ret = PCRE2_CONFIG(lib, what, buf);
	assert(ret > 0);
	return (VCL_STRING) buf;
}

VCL_INT
vmod_config_int(VRT_CTX, VCL_ENUM configs, VCL_ENUM libs)
{
	uint32_t what, where;
	int ret;
	enum code_size lib;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	lib = decode_lib_enum(libs);
	if (!have_lib[lib]) {
		VERR(ctx, "%d-bit library not supported in pcre2.config_str()",
		     code_size[lib]);
		return 0;
	}

	if (strcmp(configs, "LINKSIZE") == 0)
		what = PCRE2_CONFIG_LINKSIZE;
	else if (strcmp(configs, "MATCHLIMIT") == 0)
		what = PCRE2_CONFIG_MATCHLIMIT;
	else if (strcmp(configs, "PARENSLIMIT") == 0)
		what = PCRE2_CONFIG_PARENSLIMIT;
	else if (strcmp(configs, "RECURSIONLIMIT") == 0)
		what = PCRE2_CONFIG_RECURSIONLIMIT;
	else
		WRONG("Illegal config enum in config_int");

	ret = PCRE2_CONFIG(lib, what, &where);
	assert(ret >= 0);
	return (VCL_INT) where;
}

VCL_STRING
vmod_version(VRT_CTX __attribute__((unused)))
{
	return VERSION;
}
